"""JWT Authenticator
"""

from datetime import timedelta
import logging
from typing import Sequence

from bareasgi import WebSocketRequest, WebSocketRequestCallback

from .token_manager import TokenManager
from .types import (
    TokenStatus,
    UnauthorizedWebSocketError,
    ForbiddenWebSocketError
)

LOGGER = logging.getLogger(__name__)


class WebSocketJwtAuthenticator(TokenManager):
    """WebSocket JTW authentication middleware"""

    def __init__(
            self,
            secret: str,
            lease_expiry: timedelta,
            issuer: str,
            cookie_name: str,
            domain: str,
            path: str,
            session_expiry: timedelta,
            whitelist: Sequence[str] = ()
    ) -> None:
        """Initialise the JWT Authenticator

        Args:
            secret (str): The secret used for signing the JWT token.
            lease_expiry (timedelta): The token expiry
            issuer (str): The cookie issuer
            cookie_name (str): The cookie name
            domain (str): The cookie domain
            path (str): The cookie path
            session_expiry (timedelta): The maximum age of the cookie.
            whitelist (Sequence[str], optional): Routes for which
                authentication is not required. Defaults to None.
        """
        super().__init__(
            secret,
            lease_expiry,
            issuer,
            cookie_name,
            domain,
            path,
            session_expiry
        )
        self.whitelist = whitelist

    def _is_whitelisted(self, path: str) -> bool:
        for path_prefix in self.whitelist:
            if path.startswith(path_prefix):
                return True
        return False

    async def get_token(
            self,
            request: WebSocketRequest,
            token_status: TokenStatus
    ) -> bytes:
        """Get the token

        Args:
            request (WebSocketRequest): The request.
            token_status (TokenStatus): The token status.

        Raises:
            UnauthorizedWebSocketError: For unauthorized requests.
            ForbiddenWebSocketError: For forbidden requests.

        Returns:
            bytes: The token.
        """
        token = self.get_token_from_headers(request.scope['headers'])
        if token_status == TokenStatus.VALID:
            LOGGER.debug('Cookie still valid')
            assert token is not None
            return token

        if token_status == TokenStatus.EXPIRED and token is not None:
            raise UnauthorizedWebSocketError(request, "Expired cookie")

        LOGGER.warning('The token was invalid')
        raise ForbiddenWebSocketError(request, 'Invalid cookie')

    async def _handle_authentication(
            self,
            request: WebSocketRequest,
            handler: WebSocketRequestCallback
    ) -> None:
        LOGGER.debug("Handling authentication request")
        try:
            token = self.get_token_from_headers(request.scope['headers'])
            token_status = self.get_token_status(token)

            if token_status == TokenStatus.MISSING:
                LOGGER.debug("No token - closing websocket")
                await request.web_socket.close()
                return

            token = await self.get_token(request, token_status)

            request.context['jwt'] = self.decode(token)

            await handler(request)

        except:  # pylint: disable=bare-except
            LOGGER.exception("JWT authentication failed")
            raise

    async def _handle_whitelisted(
            self,
            request: WebSocketRequest,
            handler: WebSocketRequestCallback
    ) -> None:
        LOGGER.debug(
            'The path is whitelisted: "%s"',
            request.scope['path']
        )
        await handler(request)

    async def __call__(
            self,
            request: WebSocketRequest,
            handler: WebSocketRequestCallback
    ) -> None:

        LOGGER.debug('Jwt Authentication Request: %s', request.scope["path"])

        if self._is_whitelisted(request.scope['path']):
            await self._handle_whitelisted(request, handler)
        else:
            await self._handle_authentication(request, handler)



"""Types"""

from enum import Enum, auto
from typing import List, Optional, Tuple

from bareasgi import HttpRequest, WebSocketRequest
import bareutils.response_code as response_code

from .utils import get_host, get_scheme


class TokenStatus(Enum):
    """The token status"""
    VALID = auto()
    EXPIRED = auto()
    INVALID = auto()
    MISSING = auto()


class BareASGIHttpError(Exception):

    def __init__(
            self,
            request: HttpRequest,
            status: int,
            headers: Optional[List[Tuple[bytes, bytes]]],
            message: Optional[str]
    ) -> None:
        super().__init__(message)
        host = get_host(request.scope['headers']).decode('ascii')
        scheme = get_scheme(request.scope['headers'], request.scope['scheme'])
        self.url = f'{scheme}://{host}{request.scope["path"]}'
        self.status = status
        self.headers = headers
        self.message = message


class ForbiddenHttpError(BareASGIHttpError):

    def __init__(self, request: HttpRequest, message: str) -> None:
        super().__init__(
            request,
            response_code.FORBIDDEN,
            [(b'content_type', b'text/plain')],
            message
        )


class UnauthorizedHttpError(BareASGIHttpError):

    def __init__(self, request: HttpRequest, message: str) -> None:
        super().__init__(
            request,
            response_code.UNAUTHORIZED,
            [(b'content_type', b'text/plain')],
            message
        )


class BareASGIWebSocketError(Exception):

    def __init__(
            self,
            request: WebSocketRequest,
            status: int,
            headers: Optional[List[Tuple[bytes, bytes]]],
            message: Optional[str]
    ) -> None:
        super().__init__(message)
        host = get_host(request.scope['headers']).decode('ascii')
        scheme = get_scheme(request.scope['headers'], request.scope['scheme'])
        self.url = f'{scheme}://{host}{request.scope["path"]}'
        self.status = status
        self.headers = headers
        self.message = message


class ForbiddenWebSocketError(BareASGIWebSocketError):

    def __init__(self, request: WebSocketRequest, message: str) -> None:
        super().__init__(
            request,
            response_code.FORBIDDEN,
            [(b'content_type', b'text/plain')],
            message
        )


class UnauthorizedWebSocketError(BareASGIWebSocketError):

    def __init__(self, request: WebSocketRequest, message: str) -> None:
        super().__init__(
            request,
            response_code.UNAUTHORIZED,
            [(b'content_type', b'text/plain')],
            message
        )



"""Utilities"""

from typing import Iterable, Optional, Tuple

from bareasgi import HttpRequest, WebSocketRequest


def _find_first_header(
    names: Iterable[bytes],
    headers: Iterable[Tuple[bytes, bytes]],
    default: Optional[bytes] = None
) -> Optional[bytes]:
    for name in names:
        for key, value in headers:
            if key == name and value:
                return value
    return default


def get_host(headers: Iterable[Tuple[bytes, bytes]]) -> bytes:
    host = _find_first_header(
        (b'x-forwarded-host', b'host'),
        headers
    )
    assert host is not None
    return host


def get_scheme(headers: Iterable[Tuple[bytes, bytes]], scheme: str) -> str:
    forwarded_scheme = _find_first_header(
        (b'x-forwarded-proto',),
        headers
    )
    if forwarded_scheme is not None:
        return forwarded_scheme.decode('ascii')

    return scheme



"""JWT Authenticator
"""

from datetime import timedelta
import logging
from typing import Any, List, Mapping, Optional, Sequence, Tuple, cast
from urllib.parse import urlencode

from bareasgi import HttpRequest, HttpRequestCallback, HttpResponse, text_writer
from bareclient import HttpClient
from bareutils import header, response_code, encode_set_cookie

from .token_manager import TokenManager
from .types import (
    TokenStatus,
    BareASGIHttpError,
    UnauthorizedHttpError,
    ForbiddenHttpError
)
from .utils import get_scheme, get_host

LOGGER = logging.getLogger(__name__)


class HttpJwtAuthenticator(TokenManager):
    """JTW authentication middleware"""

    def __init__(
            self,
            secret: str,
            lease_expiry: timedelta,
            issuer: str,
            cookie_name: str,
            domain: str,
            path: str,
            session_expiry: timedelta,
            token_renewal_path: str,
            authentication_path: Optional[str] = None,
            whitelist: Sequence[str] = ()
    ) -> None:
        """Initialise the JWT Authenticator

        Args:
            secret (str): The secret used for signing the JWT token.
            lease_expiry (timedelta): The token expiry
            issuer (str): The cookie issuer
            cookie_name (str): The cookie name
            domain (str): The cookie domain
            path (str): The cookie path
            session_expiry (timedelta): The maximum age of the cookie.
            token_renewal_path (str): The path at which tokens can be renewed
            authentication_path (Optional[str], optional): The authentication
                path. Defaults to None.
            whitelist (Sequence[str], optional): Routes for which
                authentication is not required. Defaults to None.
        """
        super().__init__(
            secret,
            lease_expiry,
            issuer,
            cookie_name,
            domain,
            path,
            session_expiry
        )
        self.token_renewal_path = token_renewal_path
        self.authentication_path = authentication_path
        self.whitelist = whitelist

    async def _renew_cookie(
            self,
            request: HttpRequest,
            token: bytes
    ) -> Optional[Mapping[str, Any]]:

        host = get_host(request.scope['headers'])
        scheme = get_scheme(request.scope['headers'], request.scope['scheme'])
        base_url = f'{scheme}://{host.decode("ascii")}'

        referer_url = base_url + request.scope["path"]
        if request.scope['query_string']:
            referer_url += '?' + request.scope['query_string'].decode('utf-8')
        referer = header.find(
            b'referer',
            request.scope['headers'],
            referer_url.encode('ascii')
        )
        assert referer is not None

        headers: List[Tuple[bytes, bytes]] = [
            (b'host', host),
            (b'referer', referer),
            (b'content-length', b'0'),
            (b'connection', b'close')
        ]
        if token is not None:
            cookie = self.cookie_name + b'=' + token
            headers.append((b'cookie', cookie))

        renewal_url = base_url + self.token_renewal_path

        LOGGER.debug(
            'Renewing cookie at %s with headers %s',
            renewal_url,
            headers
        )

        async with HttpClient(
                renewal_url,
                method='POST',
                headers=headers
        ) as response:

            if response.status == response_code.NO_CONTENT:
                LOGGER.debug('Cookie renewed')
                all_set_cookies = header.set_cookie(response.headers)
                auth_set_cookies = all_set_cookies.get(
                    self.cookie_name
                )
                if auth_set_cookies is None:
                    raise RuntimeError('No cookie returned')
                return auth_set_cookies[0]

            elif response.status == response_code.UNAUTHORIZED:
                LOGGER.debug(
                    'Cookie not renewed - client requires authentication'
                )
                return None
            else:
                LOGGER.debug(
                    "Received unhandled response code %s",
                    response.status
                )

        LOGGER.debug('Cookie not renewed - failed to authenticate')
        raise UnauthorizedHttpError(request, 'Failed to authenticate')

    def _make_authenticate_location(self, request: HttpRequest) -> bytes:
        if self.authentication_path is None:
            LOGGER.debug("No path for authentication redirect")
            raise UnauthorizedHttpError(request, "Cannot authorize")

        path = request.scope['path']
        if request.scope['query_string']:
            path += '?' + request.scope['query_string'].decode()

        scheme = get_scheme(request.scope['headers'], request.scope['scheme'])
        host = get_host(request.scope['headers']).decode('ascii')
        base_url = f'{scheme}://{host}'

        redirect_url = base_url+path
        query_string = urlencode([('redirect', redirect_url)])
        location = base_url + self.authentication_path + '?' + query_string
        return location.encode()

    def _is_whitelisted(self, path: str) -> bool:
        for path_prefix in self.whitelist:
            if path.startswith(path_prefix):
                return True
        return False

    async def get_token_and_cookie(
            self,
            request: HttpRequest,
            token_status: TokenStatus
    ) -> Tuple[bytes, Optional[bytes]]:
        """Get the token and the cookie

        Args:
            request (HttpRequest): The request.
            token_status (TokenStatus): The token status.

        Raises:
            UnauthorizedHttpError: For unauthorized requests.
            ForbiddenHttpError: For forbidden requests.

        Returns:
            Tuple[bytes, Optional[bytes]]: The token and the cookie.
        """
        token = self.get_token_from_headers(request.scope['headers'])
        if token_status == TokenStatus.VALID:
            LOGGER.debug('Cookie still valid')
            assert token is not None
            return token, None

        if token_status == TokenStatus.EXPIRED and token is not None:
            LOGGER.debug('Attempting to renew cookie')

            cookie_kwargs = await self._renew_cookie(request, token)
            if cookie_kwargs is None:
                LOGGER.debug("Unable to renew cookie")
                raise UnauthorizedHttpError(request, 'Unable to renew cookie')

            token = cast(bytes, cookie_kwargs['value'])
            cookie = encode_set_cookie(**cookie_kwargs)
            return token, cookie

        LOGGER.warning('The token was invalid')
        raise ForbiddenHttpError(request, 'Invalid cookie')

    async def _handle_authentication(
            self,
            request: HttpRequest,
            handler: HttpRequestCallback
    ) -> HttpResponse:
        LOGGER.debug("Handling authentication request")
        try:
            token = self.get_token_from_headers(request.scope['headers'])
            token_status = self.get_token_status(token)

            if token_status == TokenStatus.MISSING:
                LOGGER.debug("No token - redirecting for authentication")
                headers = [
                    (b'location', self._make_authenticate_location(request))
                ]
                return HttpResponse(response_code.FOUND, headers)

            token, cookie = await self.get_token_and_cookie(request, token_status)

            request.context['jwt'] = self.decode(token)

            response = await handler(request)

            if cookie:
                if response.headers is None:
                    response.headers = []
                response.headers.append((b"set-cookie", cookie))

            return response

        except BareASGIHttpError as error:
            return HttpResponse(
                error.status,
                error.headers,
                text_writer(error.message) if error.message else None
            )
        except:  # pylint: disable=bare-except
            LOGGER.exception("JWT authentication failed")
            raise

    async def _handle_whitelisted(
            self,
            request: HttpRequest,
            handler: HttpRequestCallback
    ) -> HttpResponse:
        LOGGER.debug(
            'The path is whitelisted: "%s"',
            request.scope['path']
        )
        return await handler(request)

    async def __call__(
            self,
            request: HttpRequest,
            handler: HttpRequestCallback
    ) -> HttpResponse:

        LOGGER.debug('Jwt Authentication Request: %s', request.scope["path"])

        if self._is_whitelisted(request.scope['path']):
            return await self._handle_whitelisted(request, handler)
        else:
            return await self._handle_authentication(request, handler)



"""Token Manager"""

from datetime import datetime, timedelta
import logging
from typing import Any, Iterable, Mapping, Optional, Tuple

from bareutils import encode_set_cookie, header
import jwt

from .types import TokenStatus

LOGGER = logging.getLogger(__name__)


class TokenManager:
    """Token Manager"""

    def __init__(
            self,
            secret: str,
            lease_expiry: timedelta,
            issuer: str,
            cookie_name: str,
            domain: str,
            path: str,
            session_expiry: timedelta
    ) -> None:
        """A token manager

        Args:
            secret (str): The secret used for signing the JWT token.
            lease_expiry (timedelta): The token expiry
            issuer (str): The cookie issuer
            cookie_name (str): The cookie name
            domain (str): The cookie domain
            path (str): The cookie path
            session_expiry (timedelta): The maximum age of the cookie.
        """
        self.secret = secret
        self.lease_expiry = lease_expiry
        self.issuer = issuer
        self.cookie_name = cookie_name.encode()
        self.domain = domain.encode()
        self.path = path.encode()
        self.session_expiry = session_expiry

    def encode(
            self,
            user: str,
            now: datetime,
            issued_at: datetime,
            lease_expiry: Optional[timedelta],
            **kwargs
    ) -> bytes:
        """Encode the JSON web token.

        Args:
            user (str): The user identification
            now (datetime): The current time
            issued_at (datetime): When the token was originally issued
            lease_expiry (Optional[timedelta]): An optional expiry.

        Returns:
            str: The information encoded as a JSON web token.
        """
        if lease_expiry is None:
            lease_expiry = self.lease_expiry
        expiry = now + lease_expiry
        LOGGER.debug("Token will expire at %s", expiry)
        payload = {
            'iss': self.issuer,
            'sub': user,
            'exp': expiry,
            'iat': issued_at
        }
        payload.update(kwargs)
        token = jwt.encode(payload, key=self.secret, algorithm="HS256")
        return token.encode('ascii')

    def decode(self, token: bytes) -> Mapping[str, Any]:
        """Decode the JSON web token

        Args:
            token (bytes): The token

        Returns:
            Mapping[str, Any]: A mapping of the payload.
        """
        payload = jwt.decode(
            token.decode('ascii'),
            key=self.secret,
            options={'verify_exp': False},
            algorithms=["HS256"]
        )
        payload['exp'] = datetime.utcfromtimestamp(payload['exp'])
        payload['iat'] = datetime.utcfromtimestamp(payload['iat'])
        return payload

    def get_token_from_headers(
            self,
            headers: Iterable[Tuple[bytes, bytes]]
    ) -> Optional[bytes]:
        """Gets the token from the headers if present.

        Args:
            headers (Iterable[Tuple[bytes, bytes]]): The headers

        Returns:
            Optional[bytes]: The token or None if not found.
        """
        tokens = header.cookie(headers).get(self.cookie_name)
        if tokens is None or not tokens:
            return None
        if len(tokens) > 1:
            LOGGER.warning('Multiple tokens in header - using first')
        token = tokens[0]
        return token

    def get_jwt_payload_from_headers(
            self,
            headers: Iterable[Tuple[bytes, bytes]]
    ) -> Optional[Mapping[str, Any]]:
        """Gets the payload of the JSON web token from the headers

        Args:
            headers (Iterable[Tuple[bytes, bytes]]): The headers

        Returns:
            Optional[Mapping[str, Any]]: The payload of the JSON web token if
                present; otherwise None.
        """
        token = self.get_token_from_headers(headers)
        payload = self.decode(token) if token is not None else None
        return payload

    def generate_cookie(self, user: str) -> bytes:
        """Generate a new cookie

        Args:
            user (str): The user identification

        Returns:
            bytes: The cookie
        """
        now = datetime.utcnow()
        token = self.encode(user, now, now, None)
        return self.make_cookie(token)

    def make_cookie(self, token: bytes) -> bytes:
        """Make a cookie from a token

        Args:
            token (bytes): The token

        Returns:
            bytes: The cookie
        """
        cookie = encode_set_cookie(
            self.cookie_name,
            token,
            max_age=self.session_expiry,
            domain=self.domain,
            path=self.path,
            http_only=True
        )
        return cookie

    def get_token_status(self, token: Optional[bytes]) -> TokenStatus:
        """Get the status of the token.

        Args:
            token (Optional[bytes]): The token.

        Returns:
            TokenStatus: The status of the token.
        """
        try:
            if token is None:
                LOGGER.debug('Token missing')
                return TokenStatus.MISSING

            now = datetime.utcnow()

            payload = self.decode(token)
            expiry = payload['exp']
            if expiry < now:
                LOGGER.debug('Token expired')
                return TokenStatus.EXPIRED

            LOGGER.debug('Token valid for %s', expiry - now)
            return TokenStatus.VALID
        except:  # pylint: disable=bare-except
            LOGGER.exception('Invalid token')
            return TokenStatus.INVALID



"""bareASGI_auth_common
"""

from .helpers import add_jwt_auth_middleware
from .http_jwt_authenticator import HttpJwtAuthenticator
from .token_manager import TokenManager
from .types import (
    TokenStatus,
    BareASGIHttpError,
    ForbiddenHttpError,
    UnauthorizedHttpError,
    BareASGIWebSocketError,
    ForbiddenWebSocketError,
    UnauthorizedWebSocketError,
)
from .websocket_jwt_authenticator import WebSocketJwtAuthenticator

__all__ = [
    'add_jwt_auth_middleware',
    'HttpJwtAuthenticator',
    'TokenManager',
    'ForbiddenHttpError',
    'UnauthorizedHttpError',
    'TokenStatus',
    'BareASGIHttpError',
    'BareASGIWebSocketError',
    'ForbiddenWebSocketError',
    'UnauthorizedWebSocketError',
    'WebSocketJwtAuthenticator',
]



"""Helper functions"""

from datetime import timedelta
from typing import Optional, Sequence

from bareasgi import Application

from .http_jwt_authenticator import HttpJwtAuthenticator
from .websocket_jwt_authenticator import WebSocketJwtAuthenticator


def add_jwt_auth_middleware(
        app: Application,
        secret: str,
        lease_expiry: timedelta,
        issuer: str,
        cookie_name: str,
        domain: str,
        path: str,
        session_expiry: timedelta,
        token_renewal_path: str,
        authentication_path: Optional[str] = None,
        whitelist: Sequence[str] = ()
) -> Application:
    """Add JWT authentication middleware.

    Args:
        app (Application): The ASGI application.
        secret (str): The secret used for signing the JWT token.
        lease_expiry (timedelta): The token expiry
        issuer (str): The cookie issuer
        cookie_name (str): The cookie name
        domain (str): The cookie domain
        path (str): The cookie path
        session_expiry (timedelta): The maximum age of the cookie.
        token_renewal_path (str): The path at which tokens can be renewed
        authentication_path (Optional[str], optional): The authentication
            path. Defaults to None.
        whitelist (Sequence[str], optional): Routes for which
            authentication is not required. Defaults to None.

    Returns:
        Application: The ASGI application.
    """
    http_jwt_authenticator = HttpJwtAuthenticator(
        secret,
        lease_expiry,
        issuer,
        cookie_name,
        domain,
        path,
        session_expiry,
        token_renewal_path,
        authentication_path,
        whitelist
    )
    app.middlewares.append(http_jwt_authenticator)

    ws_jwt_authenticator = WebSocketJwtAuthenticator(
        secret,
        lease_expiry,
        issuer,
        cookie_name,
        domain,
        path,
        session_expiry,
        whitelist
    )
    app.ws_middlewares.append(ws_jwt_authenticator)

    return app



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bareasgi_auth_common']

package_data = \
{'': ['*']}

install_requires = \
['PyJWT>=2.6,<3.0', 'bareasgi>=4.4,<5.0', 'bareclient>=5.0,<6.0']

setup_kwargs = {
    'name': 'bareasgi-auth-common',
    'version': '4.1.1',
    'description': '',
    'long_description': '# bareASGI-auth-common\n\nCommon code for authentication with bareASGI.',
    'author': 'Rob Blackbourn',
    'author_email': 'rob.blackbourn@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/rob-blackbourn/bareASGI-auth-common',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
